@@ -65,7 +65,6 @@ typedef struct
  */
 static GSList *update_windows = NULL;
 static guint update_idle = 0;
-static gboolean debug_updates = FALSE;
 
 static void
 gdk_window_directfb_process_all_updates (void)
@@ -101,9 +100,7 @@ gdk_window_directfb_process_all_updates (void)
 static gboolean
 gdk_window_update_idle (gpointer data)
 {
-  GDK_THREADS_ENTER ();
   gdk_window_directfb_process_all_updates ();
-  GDK_THREADS_LEAVE ();
   
   return FALSE;
 }
@@ -116,7 +113,7 @@ gdk_window_schedule_update (GdkWindow *window)
 
   if (!update_idle)
     {
-      update_idle = g_idle_add_full (GDK_PRIORITY_REDRAW,
+      update_idle = gdk_threads_add_idle_full (GDK_PRIORITY_REDRAW,
 				     gdk_window_update_idle, NULL, NULL);
     }
 }
@@ -562,6 +559,9 @@ gdk_directfb_window_new (GdkWindow              *parent,
       gdk_directfb_event_windows_add (window);
     }
 
+  if (attributes_mask & GDK_WA_TYPE_HINT)
+    gdk_window_set_type_hint (window, attributes->type_hint);
+
   return window;
 }
 
@@ -609,23 +609,23 @@ _gdk_windowing_window_destroy (GdkWindow *window,
   if (window == gdk_directfb_focused_window)
     gdk_directfb_change_focus (NULL);
 
-/*
-printf( " DESTROYING WINDOW %p %p %d parent=%p\n",window,impl->window,recursing,(GDK_WINDOW_OBJECT (window))->parent);
-*/
+
+  if (impl->drawable.surface) {
+    GdkDrawableImplDirectFB *dimpl = GDK_DRAWABLE_IMPL_DIRECTFB (private->impl);
+    if(dimpl->cairo_surface) {
+      cairo_surface_destroy(dimpl->cairo_surface);
+      dimpl->cairo_surface= NULL;
+    }
+    impl->drawable.surface->Release (impl->drawable.surface);
+    impl->drawable.surface = NULL;
+  }
+
   if (!recursing && !foreign_destroy && impl->window ) {
     	impl->window->SetOpacity (impl->window,0);
    		impl->window->Close(impl->window);
       	impl->window->Release(impl->window);
+        impl->window = NULL;
   }
-
-#if 0 /* let the finalizer kill it */
-  if (!recursing && !foreign_destroy)
-    {
-  		if (impl->window)
-    		impl->window->Destroy (impl->window);
-  		impl->window = NULL;
-	}
-#endif
 }
 
 /* This function is called when the window is really gone.
@@ -1269,15 +1269,15 @@ _gdk_directfb_move_resize_child (GdkWindow *window,
 
   if (!private->input_only)
     {
-      if (impl->drawable.surface)
-        {
-          GdkDrawableImplDirectFB *dimpl;
-          dimpl    = GDK_DRAWABLE_IMPL_DIRECTFB (private->impl);
-          impl->drawable.surface->Release (impl->drawable.surface);
-          impl->drawable.surface = NULL;
-          cairo_surface_destroy(dimpl->cairo_surface);
-          dimpl->cairo_surface= NULL;
-        }
+    if (impl->drawable.surface) {
+      GdkDrawableImplDirectFB *dimpl = GDK_DRAWABLE_IMPL_DIRECTFB (private->impl);
+      if(dimpl->cairo_surface) {
+        cairo_surface_destroy(dimpl->cairo_surface);
+        dimpl->cairo_surface= NULL;
+      }
+    impl->drawable.surface->Release (impl->drawable.surface);
+    impl->drawable.surface = NULL;
+  }
 
       parent_impl = GDK_WINDOW_IMPL_DIRECTFB (GDK_WINDOW_OBJECT (private->parent)->impl);
 
@@ -1298,9 +1298,6 @@ _gdk_directfb_move_resize_child (GdkWindow *window,
       _gdk_directfb_move_resize_child (list->data,
                                        private->x, private->y,
                                        impl->drawable.width, impl->drawable.height);
-      //FIXEME should this really happen ?
-      if( impl->drawable.surface )
-        impl->drawable.surface->GetPosition(impl->drawable.surface,&x,&y);
     }
 }
 
@@ -1707,6 +1704,23 @@ gdk_window_set_role (GdkWindow   *window,
   /* N/A */
 }
 
+/**
+ * gdk_window_set_startup_id:
+ * @window: a toplevel #GdkWindow
+ * @startup_id: a string with startup-notification identifier
+ *
+ * When using GTK+, typically you should use gtk_window_set_startup_id()
+ * instead of this low-level function.
+ *
+ * Since: 2.12
+ *
+ **/
+void          
+gdk_window_set_startup_id (GdkWindow   *window,
+		     const gchar *startup_id)
+{
+}
+
 void
 gdk_window_set_transient_for (GdkWindow *window,
                               GdkWindow *parent)
@@ -2507,7 +2521,7 @@ gdk_window_get_frame_extents (GdkWindow    *window,
  * Given a directfb window and a subsurface of that window
  * create a gdkwindow child wrapper
  */
-#if (DIRECTFB_MICRO_VERSION > 25)
+#if (DIRECTFB_MAJOR_VERSION >= 1)
 GdkWindow *gdk_directfb_create_child_window(GdkWindow *parent,
                                 IDirectFBSurface *subsurface)
 {
@@ -2831,8 +2845,6 @@ gdk_window_impl_directfb_invalidate_maybe_recurse (GdkPaintable *paintable,
   
   if (!gdk_region_empty (visible_region))
     {
-      //if (debug_updates)
-       // draw_ugly_color (window, region);
       
       if (private->update_area)
 	{
@@ -2845,7 +2857,6 @@ gdk_window_impl_directfb_invalidate_maybe_recurse (GdkPaintable *paintable,
 	  gdk_window_schedule_update (window);
 	}
     }
-  
     gdk_region_destroy (visible_region);
 }
 
@@ -2880,14 +2891,6 @@ gdk_window_impl_directfb_process_updates (GdkPaintable *paintable,
 	  GdkRegion *expose_region;
 	  GdkRegion *window_region;
           gint width, height;
-
-          //if (debug_updates)
-           // {
-              /* Make sure we see the red invalid area before redrawing. */
-             // gdk_display_sync (gdk_drawable_get_display (window));
-              //g_usleep (70000);
-            //}
-          
 	  save_region = _gdk_windowing_window_queue_antiexpose (window, update_area);
 
 	  if (save_region)
@@ -2918,7 +2921,6 @@ gdk_window_impl_directfb_process_updates (GdkPaintable *paintable,
 	      event.expose.count = 0;
 	      event.expose.region = expose_region;
 	      gdk_region_get_clipbox (expose_region, &event.expose.area);
-	      
 	      (*_gdk_event_func) (&event, _gdk_event_data);
 	      
 	      g_object_unref (window);
@@ -2942,6 +2944,7 @@ gdk_window_impl_directfb_begin_paint_region (GdkPaintable *paintable,
   gint                     i;
 
 
+  g_assert (region != NULL );
   wimpl = GDK_WINDOW_IMPL_DIRECTFB (paintable);
   impl = (GdkDrawableImplDirectFB *)wimpl;
   impl->buffered = TRUE;
@@ -2987,11 +2990,10 @@ gdk_window_impl_directfb_end_paint (GdkPaintable *paintable)
         {
           DFBRegion reg = { impl->paint_region->extents.x1,
                             impl->paint_region->extents.y1,
-                            impl->paint_region->extents.x2 - 1,
-                            impl->paint_region->extents.y2 - 1 };
-
-          _gdk_directfb_update (impl, &reg);
+                            impl->paint_region->extents.x2 ,
+                            impl->paint_region->extents.y2 };
 
+          impl->surface->Flip(impl->surface, &reg,0);
           gdk_region_destroy (impl->paint_region);
           impl->paint_region = NULL;
         }
@@ -3009,6 +3011,41 @@ gdk_window_impl_directfb_paintable_init (GdkPaintableIface *iface)
   iface->process_updates = gdk_window_impl_directfb_process_updates;
 }
 
+void
+gdk_window_beep (GdkWindow *window)
+{
+  gdk_display_beep (gdk_display_get_default());
+}
+
+void
+gdk_window_set_opacity (GdkWindow *window,
+			gdouble    opacity)
+{
+  GdkDisplay *display;
+  guint8 cardinal;
+  
+  g_return_if_fail (GDK_IS_WINDOW (window));
+
+  if (GDK_WINDOW_DESTROYED (window))
+    return;
+
+  display = gdk_drawable_get_display (window);
+
+  if (opacity < 0)
+    opacity = 0;
+  else if (opacity > 1)
+    opacity = 1;
+  cardinal = opacity * 0xff;
+  gdk_directfb_window_set_opacity(window,cardinal);
+}
+
+void
+_gdk_windowing_window_set_composited (GdkWindow *window,
+                                      gboolean   composited)
+{
+}
+
+
 #define __GDK_WINDOW_X11_C__
 #include "gdkaliasdef.c"
 
