@@ -34,6 +34,7 @@
 #include "gtkactiongroup.h"
 #include "gtkaccellabel.h"
 #include "gtkbutton.h"
+#include "gtkiconfactory.h"
 #include "gtkimage.h"
 #include "gtkimagemenuitem.h"
 #include "gtkintl.h"
@@ -146,10 +147,11 @@ static void gtk_action_sync_tooltip     (GtkAction      *action,
 
 static GtkWidget *create_menu_item    (GtkAction *action);
 static GtkWidget *create_tool_item    (GtkAction *action);
-static void       connect_proxy       (GtkAction     *action,
-				       GtkWidget     *proxy);
+static void       connect_proxy       (GtkAction *action,
+				       GtkWidget *proxy);
 static void       disconnect_proxy    (GtkAction *action,
 				       GtkWidget *proxy);
+
 static void       closure_accel_activate (GClosure     *closure,
 					  GValue       *return_value,
 					  guint         n_param_values,
@@ -178,6 +180,7 @@ gtk_action_class_init (GtkActionClass *klass)
 
   klass->create_menu_item = create_menu_item;
   klass->create_tool_item = create_tool_item;
+  klass->create_menu = NULL;
   klass->connect_proxy = connect_proxy;
   klass->disconnect_proxy = disconnect_proxy;
 
@@ -192,6 +195,14 @@ gtk_action_class_init (GtkActionClass *klass)
 							NULL,
 							GTK_PARAM_READWRITE | 
 							G_PARAM_CONSTRUCT_ONLY));
+
+  /**
+   * GtkAction:label:
+   *
+   * The label used for menu items and buttons that activate
+   * this action. If the label is %NULL, GTK+ uses the stock 
+   * label specified via the stock-id property.
+   */
   g_object_class_install_property (gobject_class,
 				   PROP_LABEL,
 				   g_param_spec_string ("label",
@@ -223,7 +234,7 @@ gtk_action_class_init (GtkActionClass *klass)
 							NULL,
 							GTK_PARAM_READWRITE));
   /**
-   * GtkAction::icon-name:
+   * GtkAction:icon-name:
    *
    * The name of the icon from the icon theme. 
    * Note that the stock icon is preferred, if
@@ -396,7 +407,7 @@ gtk_action_new (const gchar *name,
 			 "name", name,
 			 "label", label,
 			 "tooltip", tooltip,
-			 "stock_id", stock_id,
+			 "stock-id", stock_id,
 			 NULL);
 
   return action;
@@ -682,7 +693,7 @@ connect_proxy (GtkAction     *action,
 
       if (!label)
 	label = g_object_new (GTK_TYPE_ACCEL_LABEL,
-			      "use_underline", TRUE,
+			      "use-underline", TRUE,
 			      "xalign", 0.0,
 			      "visible", TRUE,
 			      "parent", proxy,
@@ -690,7 +701,7 @@ connect_proxy (GtkAction     *action,
       
       if (GTK_IS_ACCEL_LABEL (label) && action->private_data->accel_quark)
 	g_object_set (label,
-		      "accel_closure", action->private_data->accel_closure,
+		      "accel-closure", action->private_data->accel_closure,
 		      NULL);
 
       gtk_label_set_label (GTK_LABEL (label), action->private_data->label);
@@ -712,7 +723,9 @@ connect_proxy (GtkAction     *action,
 					     image);
 	      gtk_widget_show (image);
 	    }
-	  if (action->private_data->stock_id)
+	  
+	  if (action->private_data->stock_id &&
+	      gtk_icon_factory_lookup_default (action->private_data->stock_id))
 	    gtk_image_set_from_stock (GTK_IMAGE (image),
 				      action->private_data->stock_id, GTK_ICON_SIZE_MENU);
 	  else if (action->private_data->icon_name)
@@ -728,26 +741,13 @@ connect_proxy (GtkAction     *action,
     }
   else if (GTK_IS_TOOL_ITEM (proxy))
     {
-      /* toolbar item specific synchronisers ... */
-
-      g_object_set (proxy,
-		    "visible_horizontal", action->private_data->visible_horizontal,
-		    "visible_vertical",	action->private_data->visible_vertical,
-		    "is_important", action->private_data->is_important,
-		    NULL);
-
-      gtk_action_sync_tooltip (action, proxy);
-
-      g_signal_connect_object (proxy, "create_menu_proxy",
-			       G_CALLBACK (gtk_action_create_menu_proxy),
-			       action, 0);
-
-      gtk_tool_item_rebuild_menu (GTK_TOOL_ITEM (proxy));
-
       /* toolbar button specific synchronisers ... */
       if (GTK_IS_TOOL_BUTTON (proxy))
 	{
 	  g_object_set (proxy,
+		        "visible-horizontal", action->private_data->visible_horizontal,
+		        "visible-vertical", action->private_data->visible_vertical,
+		        "is-important", action->private_data->is_important,
 			"label", action->private_data->short_label,
 			"use-underline", TRUE,
 			"stock-id", action->private_data->stock_id,
@@ -757,7 +757,23 @@ connect_proxy (GtkAction     *action,
 	  g_signal_connect_object (proxy, "clicked",
 				   G_CALLBACK (gtk_action_activate), action,
 				   G_CONNECT_SWAPPED);
-	}
+        }
+      else 
+        {
+           g_object_set (proxy,
+	 	         "visible-horizontal", action->private_data->visible_horizontal,
+		         "visible-vertical", action->private_data->visible_vertical,
+		         "is-important", action->private_data->is_important,
+		         NULL);
+       }
+
+      gtk_action_sync_tooltip (action, proxy);
+
+      g_signal_connect_object (proxy, "create_menu_proxy",
+			       G_CALLBACK (gtk_action_create_menu_proxy),
+			       action, 0);
+
+      gtk_tool_item_rebuild_menu (GTK_TOOL_ITEM (proxy));
     }
   else if (GTK_IS_BUTTON (proxy))
     {
@@ -771,17 +787,13 @@ connect_proxy (GtkAction     *action,
 	}
       else 
 	{
-	  GtkWidget *image;
-
-	  image = gtk_button_get_image (GTK_BUTTON (proxy));
-
 	  if (GTK_BIN (proxy)->child == NULL || 
 	      GTK_IS_LABEL (GTK_BIN (proxy)->child))
 	    {
 	      /* synchronise the label */
 	      g_object_set (proxy,
 			    "label", action->private_data->short_label,
-			    "use_underline", TRUE,
+			    "use-underline", TRUE,
 			    NULL);
 	    }
 	}
@@ -1332,7 +1344,7 @@ gtk_action_set_short_label (GtkAction	*action,
 
       if (GTK_IS_TOOL_BUTTON (proxy))
 	gtk_tool_button_set_label (GTK_TOOL_BUTTON (proxy), 
-				   action->private_data->label);
+				   action->private_data->short_label);
       else if (GTK_IS_BUTTON (proxy) &&
 	       !gtk_button_get_use_stock (GTK_BUTTON (proxy)))
 	{
@@ -1340,7 +1352,7 @@ gtk_action_set_short_label (GtkAction	*action,
 	  
 	  if (child == NULL || GTK_IS_LABEL (child))
 	    gtk_button_set_label (GTK_BUTTON (proxy), 
-				  action->private_data->label);
+				  action->private_data->short_label);
 	}
     }
 
@@ -1524,8 +1536,6 @@ gtk_action_set_icon_name (GtkAction   *action,
       else if (GTK_IS_BUTTON (proxy) &&
 	       !gtk_button_get_use_stock (GTK_BUTTON (proxy)))
 	{
-	  GtkWidget *image;
-
 	  image = gtk_button_get_image (GTK_BUTTON (proxy));
 	  
 	  if (GTK_IS_IMAGE (image) &&
@@ -1766,5 +1776,28 @@ gtk_action_disconnect_accelerator (GtkAction *action)
 				action->private_data->accel_closure);
 }
 
+/**
+ * gtk_action_create_menu:
+ * @action: a #GtkAction
+ *
+ * If @action provides a #GtkMenu widget as a submenu for the menu
+ * item or the toolbar item it creates, this function returns an
+ * instance of that menu.
+ *
+ * Return value: the menu item provided by the action, or %NULL.
+ *
+ * Since: 2.12
+ */
+GtkWidget *
+gtk_action_create_menu (GtkAction *action)
+{
+  g_return_val_if_fail (GTK_IS_ACTION (action), NULL);
+
+  if (GTK_ACTION_GET_CLASS (action)->create_menu)
+    return GTK_ACTION_GET_CLASS (action)->create_menu (action);
+
+  return NULL;
+}
+
 #define __GTK_ACTION_C__
 #include "gtkaliasdef.c"
