@@ -97,7 +97,7 @@ enum
   PROP_STORAGE_TYPE
 };
 
-G_DEFINE_TYPE (GtkImage, gtk_image, GTK_TYPE_MISC);
+G_DEFINE_TYPE (GtkImage, gtk_image, GTK_TYPE_MISC)
 
 static void
 gtk_image_class_init (GtkImageClass *class)
@@ -192,8 +192,8 @@ gtk_image_class_init (GtkImageClass *class)
   /**
    * GtkImage:pixel-size:
    *
-   * The :pixel-size property can be used to specify a fixed size
-   * overriding the :icon-size property for images of type 
+   * The "pixel-size" property can be used to specify a fixed size
+   * overriding the #GtkImage:icon-size property for images of type 
    * %GTK_IMAGE_ICON_NAME. 
    *
    * Since: 2.6
@@ -218,7 +218,7 @@ gtk_image_class_init (GtkImageClass *class)
   /**
    * GtkImage:icon-name:
    *
-   * The name of the icon in the icon theme.  If the icon theme is
+   * The name of the icon in the icon theme. If the icon theme is
    * changed, the image will be updated automatically.
    *
    * Since: 2.6
@@ -489,8 +489,7 @@ gtk_image_new_from_pixmap (GdkPixmap *pixmap,
  * 
  * Creates a #GtkImage widget displaying a @image with a @mask.
  * A #GdkImage is a client-side image buffer in the pixel format of the
- * current display.
- * The #GtkImage does not assume a reference to the
+ * current display. The #GtkImage does not assume a reference to the
  * image or mask; you still need to unref them if you own references.
  * #GtkImage will add its own reference rather than adopting yours.
  * 
@@ -553,9 +552,9 @@ gtk_image_new_from_file   (const gchar *filename)
  * pixbuf; you still need to unref it if you own references.
  * #GtkImage will add its own reference rather than adopting yours.
  * 
- * Note that this function just creates an #GtkImage from the pixbuf.  The
- * #GtkImage created will not react to state changes.  Should you want that, you
- * should use gtk_image_new_from_icon_set().
+ * Note that this function just creates an #GtkImage from the pixbuf. The
+ * #GtkImage created will not react to state changes. Should you want that, 
+ * you should use gtk_image_new_from_icon_set().
  * 
  * Return value: a new #GtkImage
  **/
@@ -615,7 +614,6 @@ gtk_image_new_from_stock (const gchar    *stock_id,
  * icon set; you still need to unref it if you own references.
  * #GtkImage will add its own reference rather than adopting yours.
  * 
- * 
  * Return value: a new #GtkImage
  **/
 GtkWidget*
@@ -695,7 +693,6 @@ gtk_image_new_from_icon_name (const gchar    *icon_name,
  * @mask: a #GdkBitmap or %NULL
  *
  * See gtk_image_new_from_pixmap() for details.
- * 
  **/
 void
 gtk_image_set_from_pixmap (GtkImage  *image,
@@ -747,7 +744,6 @@ gtk_image_set_from_pixmap (GtkImage  *image,
  * @mask: a #GdkBitmap or %NULL
  *
  * See gtk_image_new_from_image() for details.
- * 
  **/
 void
 gtk_image_set_from_image  (GtkImage  *image,
@@ -798,7 +794,6 @@ gtk_image_set_from_image  (GtkImage  *image,
  * @filename: a filename or %NULL
  *
  * See gtk_image_new_from_file() for details.
- * 
  **/
 void
 gtk_image_set_from_file   (GtkImage    *image,
@@ -855,7 +850,6 @@ gtk_image_set_from_file   (GtkImage    *image,
  * @pixbuf: a #GdkPixbuf or %NULL
  *
  * See gtk_image_new_from_pixbuf() for details. 
- * 
  **/
 void
 gtk_image_set_from_pixbuf (GtkImage  *image,
@@ -895,7 +889,6 @@ gtk_image_set_from_pixbuf (GtkImage  *image,
  * @size: a stock icon size
  *
  * See gtk_image_new_from_stock() for details.
- * 
  **/
 void
 gtk_image_set_from_stock  (GtkImage       *image,
@@ -939,7 +932,6 @@ gtk_image_set_from_stock  (GtkImage       *image,
  * @size: a stock icon size
  *
  * See gtk_image_new_from_icon_set() for details.
- * 
  **/
 void
 gtk_image_set_from_icon_set  (GtkImage       *image,
@@ -1088,7 +1080,6 @@ gtk_image_get_storage_type (GtkImage *image)
  * %GTK_IMAGE_PIXMAP (see gtk_image_get_storage_type()).
  * The caller of this function does not own a reference to the
  * returned pixmap and mask.
- * 
  **/
 void
 gtk_image_get_pixmap (GtkImage   *image,
@@ -1138,7 +1129,6 @@ gtk_image_get_image  (GtkImage   *image,
  * gtk_image_get_pixbuf:
  * @image: a #GtkImage
  *
- *
  * Gets the #GdkPixbuf being displayed by the #GtkImage.
  * The storage type of the image must be %GTK_IMAGE_EMPTY or
  * %GTK_IMAGE_PIXBUF (see gtk_image_get_storage_type()).
@@ -1171,7 +1161,6 @@ gtk_image_get_pixbuf (GtkImage *image)
  * %GTK_IMAGE_STOCK (see gtk_image_get_storage_type()).
  * The returned string is owned by the #GtkImage and should not
  * be freed.
- * 
  **/
 void
 gtk_image_get_stock  (GtkImage        *image,
@@ -1201,7 +1190,6 @@ gtk_image_get_stock  (GtkImage        *image,
  * Gets the icon set and size being displayed by the #GtkImage.
  * The storage type of the image must be %GTK_IMAGE_EMPTY or
  * %GTK_IMAGE_ICON_SET (see gtk_image_get_storage_type()).
- * 
  **/
 void
 gtk_image_get_icon_set  (GtkImage        *image,
@@ -1223,7 +1211,6 @@ gtk_image_get_icon_set  (GtkImage        *image,
  * gtk_image_get_animation:
  * @image: a #GtkImage
  *
- *
  * Gets the #GdkPixbufAnimation being displayed by the #GtkImage.
  * The storage type of the image must be %GTK_IMAGE_EMPTY or
  * %GTK_IMAGE_ANIMATION (see gtk_image_get_storage_type()).
@@ -1356,8 +1343,6 @@ animation_timeout (gpointer data)
 {
   GtkImage *image;
 
-  GDK_THREADS_ENTER ();
-
   image = GTK_IMAGE (data);
   
   image->data.anim.frame_timeout = 0;
@@ -1366,7 +1351,7 @@ animation_timeout (gpointer data)
 
   if (gdk_pixbuf_animation_iter_get_delay_time (image->data.anim.iter) >= 0)
     image->data.anim.frame_timeout =
-      g_timeout_add (gdk_pixbuf_animation_iter_get_delay_time (image->data.anim.iter),
+      gdk_threads_add_timeout (gdk_pixbuf_animation_iter_get_delay_time (image->data.anim.iter),
                      animation_timeout,
                      image);
 
@@ -1375,19 +1360,20 @@ animation_timeout (gpointer data)
   if (GTK_WIDGET_DRAWABLE (image))
     gdk_window_process_updates (GTK_WIDGET (image)->window, TRUE);
 
-  GDK_THREADS_LEAVE ();
-
   return FALSE;
 }
 
 static void
 icon_theme_changed (GtkImage *image)
 {
-  if (image->data.name.pixbuf)
-    g_object_unref (image->data.name.pixbuf);
-  image->data.name.pixbuf = NULL;
+  if (image->storage_type == GTK_IMAGE_ICON_NAME) 
+    {
+      if (image->data.name.pixbuf)
+	g_object_unref (image->data.name.pixbuf);
+      image->data.name.pixbuf = NULL;
 
-  gtk_widget_queue_draw (GTK_WIDGET (image));
+      gtk_widget_queue_draw (GTK_WIDGET (image));
+    }
 }
 
 static void
@@ -1674,7 +1660,7 @@ gtk_image_expose (GtkWidget      *widget,
                 
                 if (gdk_pixbuf_animation_iter_get_delay_time (image->data.anim.iter) >= 0)
                   image->data.anim.frame_timeout =
-                    g_timeout_add (gdk_pixbuf_animation_iter_get_delay_time (image->data.anim.iter),
+                    gdk_threads_add_timeout (gdk_pixbuf_animation_iter_get_delay_time (image->data.anim.iter),
                                    animation_timeout,
                                    image);
               }
@@ -1895,8 +1881,7 @@ gtk_image_reset (GtkImage *image)
       break;
 
     case GTK_IMAGE_ICON_NAME:
-      if (image->data.name.icon_name)
-	g_free (image->data.name.icon_name);
+      g_free (image->data.name.icon_name);
       image->data.name.icon_name = NULL;
       if (image->data.name.pixbuf)
 	g_object_unref (image->data.name.pixbuf);
