@@ -64,6 +64,7 @@ enum {
   PROP_DOUBLE_CLICK_DISTANCE,
   PROP_CURSOR_BLINK,
   PROP_CURSOR_BLINK_TIME,
+  PROP_CURSOR_BLINK_TIMEOUT,
   PROP_SPLIT_CURSOR,
   PROP_THEME_NAME,
   PROP_ICON_THEME_NAME,
@@ -84,6 +85,7 @@ enum {
   PROP_CURSOR_THEME_SIZE,
 #endif
   PROP_ALTERNATIVE_BUTTON_ORDER,
+  PROP_ALTERNATIVE_SORT_ARROWS,
   PROP_SHOW_INPUT_METHOD_MENU,
   PROP_SHOW_UNICODE_MENU,
   PROP_TIMEOUT_INITIAL,
@@ -92,7 +94,18 @@ enum {
   PROP_COLOR_SCHEME,
   PROP_ENABLE_ANIMATIONS,
   PROP_TOUCHSCREEN_MODE,
-  PROP_COLOR_HASH
+  PROP_TOOLTIP_TIMEOUT,
+  PROP_TOOLTIP_BROWSE_TIMEOUT,
+  PROP_TOOLTIP_BROWSE_MODE_TIMEOUT,
+  PROP_KEYNAV_CURSOR_ONLY,
+  PROP_KEYNAV_WRAP_AROUND,
+  PROP_ERROR_BELL,
+  PROP_COLOR_HASH,
+  PROP_FILE_CHOOSER_BACKEND,
+  PROP_PRINT_BACKENDS,
+  PROP_PRINT_PREVIEW_COMMAND,
+  PROP_ENABLE_MNEMONICS,
+  PROP_ENABLE_ACCELS
 };
 
 
@@ -122,7 +135,7 @@ static void    settings_update_font_options      (GtkSettings           *setting
 static void    settings_update_color_scheme      (GtkSettings *settings);
 
 static void    merge_color_scheme                (GtkSettings           *settings, 
-						  GValue                *value, 
+						  const GValue          *value, 
 						  GtkSettingsSource      source);
 static gchar  *get_color_scheme                  (GtkSettings           *settings);
 static GHashTable *get_color_hash                (GtkSettings           *settings);
@@ -201,6 +214,15 @@ gtk_settings_class_init (GtkSettingsClass *class)
                                                                GTK_PARAM_READWRITE),
                                              NULL);
   g_assert (result == PROP_DOUBLE_CLICK_DISTANCE);
+
+  /**
+   * GtkSettings:gtk-cursor-blink:
+   *
+   * Whether the cursor should blink. 
+   *
+   * Also see the #GtkSettings:gtk-cursor-blink-timeout setting, 
+   * which allows more flexible control over cursor blinking.
+   */
   result = settings_install_property_parser (class,
                                              g_param_spec_boolean ("gtk-cursor-blink",
 								   P_("Cursor Blink"),
@@ -212,11 +234,31 @@ gtk_settings_class_init (GtkSettingsClass *class)
   result = settings_install_property_parser (class,
                                              g_param_spec_int ("gtk-cursor-blink-time",
                                                                P_("Cursor Blink Time"),
-                                                               P_("Length of the cursor blink cycle, in milleseconds"),
+                                                               P_("Length of the cursor blink cycle, in milliseconds"),
                                                                100, G_MAXINT, 1200,
                                                                GTK_PARAM_READWRITE),
                                              NULL);
   g_assert (result == PROP_CURSOR_BLINK_TIME);
+ 
+  /**
+   * GtkSettings:gtk-cursor-blink-timeout:
+   *
+   * Time after which the cursor stops blinking, in seconds.
+   * The timer is reset after each user interaction.
+   *
+   * Setting this to zero has the same effect as setting
+   * #GtkSettings:gtk-cursor-blinks to %FALSE. 
+   *
+   * Since: 2.12
+   */
+  result = settings_install_property_parser (class,
+                                             g_param_spec_int ("gtk-cursor-blink-timeout",
+                                                               P_("Cursor Blink Timeout"),
+                                                               P_("Time after which the cursor stops blinking, in seconds"),
+                                                               1, G_MAXINT, G_MAXINT,
+                                                               GTK_PARAM_READWRITE),
+                                             NULL);
+  g_assert (result == PROP_CURSOR_BLINK_TIMEOUT);
   result = settings_install_property_parser (class,
                                              g_param_spec_boolean ("gtk-split-cursor",
 								   P_("Split Cursor"),
@@ -386,6 +428,24 @@ gtk_settings_class_init (GtkSettingsClass *class)
                                              NULL);
   g_assert (result == PROP_ALTERNATIVE_BUTTON_ORDER);
 
+  /**
+   * GtkSettings:gtk-alternative-sort-arrows:
+   *
+   * Controls the direction of the sort indicators in sorted list and tree
+   * views. By default an arrow pointing down means the column is sorted
+   * in ascending order. When set to %TRUE, this order will be inverted.
+   *
+   * Since: 2.12
+   */
+  result = settings_install_property_parser (class,
+                                             g_param_spec_boolean ("gtk-alternative-sort-arrows",
+								   P_("Alternative sort indicator direction"),
+								   P_("Whether the direction of the sort indicators in list and tree views is inverted compared to the default (where down means ascending)"),
+								   FALSE,
+								   GTK_PARAM_READWRITE),
+                                             NULL);
+  g_assert (result == PROP_ALTERNATIVE_SORT_ARROWS);
+
   result = settings_install_property_parser (class,
 					     g_param_spec_boolean ("gtk-show-input-method-menu",
 								   P_("Show the 'Input Methods' menu"),
@@ -452,6 +512,12 @@ gtk_settings_class_init (GtkSettingsClass *class)
    * merged, color specifications will be converted to hexadecimal form
    * when getting this property.
    *
+   * Starting with GTK+ 2.12, the entries can alternatively be separated
+   * by ';' instead of newlines:
+   * <programlisting>
+   * name1: color1; name2: color2; ...
+   * </programlisting>
+   *
    * Since: 2.10
    */
   result = settings_install_property_parser (class,
@@ -477,7 +543,7 @@ gtk_settings_class_init (GtkSettingsClass *class)
   /**
    * GtkSettings:gtk-touchscreen-mode:
    *
-   * When TRUE, there are no motion notify events delivered on this screen,
+   * When %TRUE, there are no motion notify events delivered on this screen,
    * and widgets can't use the pointer hovering them for any essential
    * functionality.
    *
@@ -493,23 +559,232 @@ gtk_settings_class_init (GtkSettingsClass *class)
 
   g_assert (result == PROP_TOUCHSCREEN_MODE);
 
+  /**
+   * GtkSettings:gtk-tooltip-timeout:
+   *
+   * Time, in milliseconds, after which a tooltip could appear if the
+   * cursor is hovering on top of a widget.
+   *
+   * Since: 2.12
+   */
+  result = settings_install_property_parser (class,
+					     g_param_spec_int ("gtk-tooltip-timeout",
+ 							       P_("Tooltip timeout"),
+ 							       P_("Timeout before tooltip is shown"),
+ 							       0, G_MAXINT,
+							       1500,
+ 							       GTK_PARAM_READWRITE),
+					     NULL);
+
+  g_assert (result == PROP_TOOLTIP_TIMEOUT);
+
+  /**
+   * GtkSettings:gtk-tooltip-browse-timeout:
+   *
+   * Controls the time after which tooltips will appear when
+   * browse mode is enabled, in milliseconds.
+   *
+   * Browse mode is enabled when the mouse pointer moves off an object
+   * where a tooltip was currently being displayed. If the mouse pointer
+   * hits another object before the browse mode timeout expires (see
+   * #GtkSettings:gtk-tooltip-browse-mode-timeout), it will take the 
+   * amount of milliseconds specified by this setting to popup the tooltip
+   * for the new object.
+   *
+   * Since: 2.12
+   */
+  result = settings_install_property_parser (class,
+					     g_param_spec_int ("gtk-tooltip-browse-timeout",
+ 							       P_("Tooltip browse timeout"),
+ 							       P_("Timeout before tooltip is shown when browse mode is enabled"),
+ 							       0, G_MAXINT,
+							       100,
+ 							       GTK_PARAM_READWRITE),
+					     NULL);
+
+  g_assert (result == PROP_TOOLTIP_BROWSE_TIMEOUT);
+
+  /**
+   * GtkSettings:gtk-tooltip-browse-mode-timeout:
+   *
+   * Amount of time, in milliseconds, after which the browse mode
+   * will be disabled.
+   *
+   * See #GtkSettings:gtk-tooltip-browse-timeout for more information
+   * about browse mode.
+   *
+   * Since: 2.12
+   */
+  result = settings_install_property_parser (class,
+					     g_param_spec_int ("gtk-tooltip-browse-mode-timeout",
+ 							       P_("Tooltip browse mode timeout"),
+ 							       P_("Timeout after which browse mode is disabled"),
+ 							       0, G_MAXINT,
+							       500,
+ 							       GTK_PARAM_READWRITE),
+					     NULL);
+
+  g_assert (result == PROP_TOOLTIP_BROWSE_MODE_TIMEOUT);
+
+  /**
+   * GtkSettings:gtk-keynav-cursor-only:
+   *
+   * When %TRUE, keyboard navigation should be able to reach all widgets
+   * by using the cursor keys only. Tab, Shift etc. keys can't be expected
+   * to be present on the used input device.
+   *
+   * Since: 2.12
+   */
+  result = settings_install_property_parser (class,
+                                             g_param_spec_boolean ("gtk-keynav-cursor-only",
+                                                                   P_("Keynav Cursor Only"),
+                                                                   P_("When TRUE, there are only cursor keys available to navigate widgets"),
+                                                                   FALSE,
+                                                                   GTK_PARAM_READWRITE),
+                                             NULL);
+
+  g_assert (result == PROP_KEYNAV_CURSOR_ONLY);
+
+  /**
+   * GtkSettings:gtk-keynav-wrap-around:
+   *
+   * When %TRUE, some widgets will wrap around when doing keyboard
+   * navigation, such as menus, menubars and notebooks.
+   *
+   * Since: 2.12
+   */
+  result = settings_install_property_parser (class,
+                                             g_param_spec_boolean ("gtk-keynav-wrap-around",
+                                                                   P_("Keynav Wrap Around"),
+                                                                   P_("Whether to wrap around when keyboard-navigating widgets"),
+                                                                   TRUE,
+                                                                   GTK_PARAM_READWRITE),
+                                             NULL);
+
+  g_assert (result == PROP_KEYNAV_WRAP_AROUND);
+
+  /**
+   * GtkSettings:gtk-error-bell:
+   *
+   * When %TRUE, keyboard navigation and other input-related errors
+   * will cause a beep. Since the error bell is implemented using
+   * gdk_window_beep(), the windowing system may offer ways to
+   * configure the error bell in many ways, such as flashing the
+   * window or similar visual effects.
+   *
+   * Since: 2.12
+   */
+  result = settings_install_property_parser (class,
+                                             g_param_spec_boolean ("gtk-error-bell",
+                                                                   P_("Error Bell"),
+                                                                   P_("When TRUE, keyboard navigation and other errors will cause a beep"),
+                                                                   TRUE,
+                                                                   GTK_PARAM_READWRITE),
+                                             NULL);
+
+  g_assert (result == PROP_ERROR_BELL);
+
   /**
    * GtkSettings:color-hash:
    *
-   * Holds a hash table representation of the gtk-color-scheme setting,
-   * mapping color names to #GdkColor<!-- -->s. 
+   * Holds a hash table representation of the #GtkSettings:gtk-color-scheme 
+   * setting, mapping color names to #GdkColor<!-- -->s. 
    *
    * Since: 2.10
    */
-  g_object_class_install_property (gobject_class,
-				   PROP_COLOR_HASH,
-				   g_param_spec_boxed ("color-hash",
-						       P_("Color Hash"),
-						       P_("A hash table representation of the color scheme."),
-						       G_TYPE_HASH_TABLE,
-						       GTK_PARAM_READABLE));
+  result = settings_install_property_parser (class, 
+                                             g_param_spec_boxed ("color-hash",
+                                                                 P_("Color Hash"),
+                                                                 P_("A hash table representation of the color scheme."),
+                                                                 G_TYPE_HASH_TABLE,
+                                                                 GTK_PARAM_READABLE),
+                                             NULL);
+  g_assert (result == PROP_COLOR_HASH);
 
-  class_n_properties++;
+  result = settings_install_property_parser (class, 
+                                             g_param_spec_string ("gtk-file-chooser-backend",
+                                                                  P_("Default file chooser backend"),
+                                                                  P_("Name of the GtkFileChooser backend to use by default"),
+                                                                  NULL,
+                                                                  GTK_PARAM_READWRITE),
+                                             NULL);
+  g_assert (result == PROP_FILE_CHOOSER_BACKEND);
+
+  /**
+   * GtkSettings:gtk-print-backends:
+   *
+   * A comma-separated list of print backends to use in the print
+   * dialog. Available print backends depend on the GTK+ installation,
+   * and may include "pdf", "cups" or "lpr".
+   *
+   * Since: 2.10
+   */
+  result = settings_install_property_parser (class,
+                                             g_param_spec_string ("gtk-print-backends",
+                                                                  P_("Default print backend"),
+                                                                  P_("List of the GtkPrintBackend backends to use by default"),
+                                                                  GTK_PRINT_BACKENDS,
+                                                                  GTK_PARAM_READWRITE),
+                                             NULL);
+  g_assert (result == PROP_PRINT_BACKENDS);
+
+  /**
+   * GtkSettings:gtk-print-preview-command:
+   *
+   * A command to run for displaying the print preview. The command
+   * should contain a %f placeholder, which will get replaced by
+   * the path to the pdf file. The command may also contain a %s
+   * placeholder, which will get replaced by the path to a file
+   * containing the print settings in the format produced by 
+   * gtk_print_settings_to_file().
+   *
+   * The preview application is responsible for removing the pdf file
+   * and the print settings file when it is done.
+   *
+   * Since: 2.10
+   */
+  result = settings_install_property_parser (class,
+                                             g_param_spec_string ("gtk-print-preview-command",
+                                                                  P_("Default command to run when displaying a print preview"),
+                                                                  P_("Command to run when displaying a print preview"),
+                                                                  GTK_PRINT_PREVIEW_COMMAND,
+                                                                  GTK_PARAM_READWRITE),
+                                             NULL); 
+  g_assert (result == PROP_PRINT_PREVIEW_COMMAND);
+
+  /**
+   * GtkSettings:gtk-enable-mnemonics:
+   *
+   * Whether labels and menu items should have visible mnemonics which
+   * can be activated.
+   *
+   * Since: 2.12
+   */
+  result = settings_install_property_parser (class,
+                                             g_param_spec_boolean ("gtk-enable-mnemonics",
+                                                                   P_("Enable Mnemonics"),
+                                                                   P_("Whether labels should have mnemonics"),
+                                                                   TRUE,
+                                                                   GTK_PARAM_READWRITE),
+                                             NULL);
+  g_assert (result == PROP_ENABLE_MNEMONICS);
+
+  /**
+   * GtkSettings:gtk-enable-accels:
+   *
+   * Whether menu items should have visible accelerators which can be
+   * activated.
+   *
+   * Since: 2.12
+   */
+  result = settings_install_property_parser (class,
+                                             g_param_spec_boolean ("gtk-enable-accels",
+                                                                   P_("Enable Accelerators"),
+                                                                   P_("Whether menu items should have accelerators"),
+                                                                   TRUE,
+                                                                   GTK_PARAM_READWRITE),
+                                             NULL);
+  g_assert (result == PROP_ENABLE_ACCELS);
 }
 
 static void
@@ -599,6 +874,9 @@ gtk_settings_set_property (GObject      *object,
 
   g_value_copy (value, &settings->property_values[property_id - 1].value);
   settings->property_values[property_id - 1].source = GTK_SETTINGS_SOURCE_APPLICATION;
+  
+  if (pspec->param_id == PROP_COLOR_SCHEME)
+    merge_color_scheme (settings, value, GTK_SETTINGS_SOURCE_APPLICATION);
 }
 
 static void
@@ -611,10 +889,16 @@ gtk_settings_get_property (GObject     *object,
   GType value_type = G_VALUE_TYPE (value);
   GType fundamental_type = G_TYPE_FUNDAMENTAL (value_type);
 
-  if (property_id == PROP_COLOR_HASH)
+  /* handle internal properties */
+  switch (property_id)
     {
+    case PROP_COLOR_HASH:
       g_value_set_boxed (value, get_color_hash (settings));
       return;
+    case PROP_COLOR_SCHEME:
+      g_value_take_string (value, get_color_scheme (settings));
+      return;
+    default: ;
     }
 
   /* For enums and strings, we need to get the value as a string,
@@ -630,15 +914,7 @@ gtk_settings_get_property (GObject     *object,
 	  !gdk_screen_get_setting (settings->screen, pspec->name, value))
         g_value_copy (&settings->property_values[property_id - 1].value, value);
       else 
-	{
-	  if (pspec->param_id == PROP_COLOR_SCHEME)
-	    {
-	      merge_color_scheme (settings, value, GTK_SETTINGS_SOURCE_XSETTING);
-	      g_value_take_string (value, get_color_scheme (settings));
-	    }
-	  
-	  g_param_value_validate (pspec, value);
-	}
+        g_param_value_validate (pspec, value);
     }
   else
     {
@@ -809,18 +1085,15 @@ apply_queued_setting (GtkSettings             *data,
   if (_gtk_settings_parse_convert (parser, &qvalue->public.value,
 				   pspec, &tmp_value))
     {
-      if (pspec->param_id == PROP_COLOR_SCHEME) 
+      if (data->property_values[pspec->param_id - 1].source <= qvalue->source)
 	{
-	  merge_color_scheme (data, &tmp_value, qvalue->source);
-	  g_object_set_property (G_OBJECT (data), pspec->name, &tmp_value);
-	  data->property_values[pspec->param_id - 1].source = GTK_SETTINGS_SOURCE_DEFAULT;
-	}      
-      
-      else if (data->property_values[pspec->param_id - 1].source <= qvalue->source)
-	{
-	  g_object_set_property (G_OBJECT (data), pspec->name, &tmp_value);
+          g_value_copy (&tmp_value, &data->property_values[pspec->param_id - 1].value);
 	  data->property_values[pspec->param_id - 1].source = qvalue->source;
+          g_object_notify (G_OBJECT (data), g_param_spec_get_name (pspec));
 	}
+
+      if (pspec->param_id == PROP_COLOR_SCHEME) 
+        merge_color_scheme (data, &tmp_value, qvalue->source);
     }
   else
     {
@@ -857,6 +1130,12 @@ settings_install_property_parser (GtkSettingsClass   *class,
     case G_TYPE_DOUBLE:
     case G_TYPE_STRING:
       break;
+    case G_TYPE_BOXED:
+      if (strcmp (g_param_spec_get_name (pspec), "color-hash") == 0)
+        {
+          break;
+        }
+      /* fall through */
     default:
       if (!parser)
         {
@@ -1412,10 +1691,30 @@ gtk_rc_property_parse_border (const GParamSpec *pspec,
 void
 _gtk_settings_handle_event (GdkEventSetting *event)
 {
-  GtkSettings *settings = gtk_settings_get_for_screen (gdk_drawable_get_screen (event->window));
-  
-  if (g_object_class_find_property (G_OBJECT_GET_CLASS (settings), event->name))
-    g_object_notify (G_OBJECT (settings), event->name);
+  GtkSettings *settings;
+  GParamSpec *pspec;
+  guint property_id;
+
+  settings = gtk_settings_get_for_screen (gdk_drawable_get_screen (event->window));
+  pspec = g_object_class_find_property (G_OBJECT_GET_CLASS (settings), event->name);
+ 
+  if (pspec) 
+    {
+      property_id = pspec->param_id;
+
+      if (property_id == PROP_COLOR_SCHEME)
+        {
+          GValue value = { 0, };
+
+          g_value_init (&value, G_TYPE_STRING);
+          if (!gdk_screen_get_setting (settings->screen, pspec->name, &value))
+            g_value_set_static_string (&value, "");
+          merge_color_scheme (settings, &value, GTK_SETTINGS_SOURCE_XSETTING);
+          g_value_unset (&value);
+        }
+
+      g_object_notify (G_OBJECT (settings), pspec->name);
+   }
 }
 
 static void
@@ -1545,10 +1844,6 @@ settings_update_font_options (GtkSettings *settings)
 
   options = cairo_font_options_create ();
   
-  /* hint_metrics = FALSE should never be set for user interface code.
-   */
-  cairo_font_options_set_hint_metrics (options, CAIRO_HINT_METRICS_ON);
-
   if (hinting >= 0 && !hinting)
     {
       hint_style = CAIRO_HINT_STYLE_NONE;
@@ -1565,8 +1860,7 @@ settings_update_font_options (GtkSettings *settings)
 	hint_style = CAIRO_HINT_STYLE_FULL;
     }
 
-  if (hint_style_str)
-    g_free (hint_style_str);
+  g_free (hint_style_str);
 
   cairo_font_options_set_hint_style (options, hint_style);
 
@@ -1619,40 +1913,63 @@ settings_update_resolution (GtkSettings *settings)
 }
 #endif
 
-static void
-settings_update_color_scheme (GtkSettings *settings)
-{
-  gchar *dummy;
-
-  g_object_get (settings, "gtk-color-scheme", &dummy, NULL);
-
-  /* nothing to do here, the color hash is updated as a
-   * side effect of getting the color scheme
-   */
-
-  g_free (dummy);  
-}
-
-
 typedef struct {
   GHashTable *color_hash;
   GHashTable *tables[GTK_SETTINGS_SOURCE_APPLICATION + 1];
   gchar *lastentry[GTK_SETTINGS_SOURCE_APPLICATION + 1];
 } ColorSchemeData;
 
+static void
+color_scheme_data_free (ColorSchemeData *data)
+{
+  gint i;
+
+  g_hash_table_unref (data->color_hash);
+
+  for (i = 0; i <= GTK_SETTINGS_SOURCE_APPLICATION; i++)
+    {
+      if (data->tables[i])
+	g_hash_table_unref (data->tables[i]);
+      g_free (data->lastentry[i]);
+    }
+
+  g_free (data);
+}
+
+static void
+settings_update_color_scheme (GtkSettings *settings)
+{
+  if (!g_object_get_data (G_OBJECT (settings), "gtk-color-scheme"))
+    {
+      ColorSchemeData *data;
+      GValue value = { 0, };
+
+      data = g_new0 (ColorSchemeData, 1);
+      data->color_hash = g_hash_table_new_full (g_str_hash, g_str_equal, g_free,
+					        (GDestroyNotify) gdk_color_free);
+      g_object_set_data_full (G_OBJECT (settings), "gtk-color-scheme",
+			      data, (GDestroyNotify) color_scheme_data_free); 
+
+      g_value_init (&value, G_TYPE_STRING);
+      if (gdk_screen_get_setting (settings->screen, "gtk-color-scheme", &value))
+        {
+          merge_color_scheme (settings, &value, GTK_SETTINGS_SOURCE_XSETTING);
+          g_value_unset (&value);
+        }
+   }
+}
 
 static gboolean
 add_color_to_hash (gchar      *name, 
 		   GdkColor   *color, 
 		   GHashTable *target)
 {
-  GdkColor *old, *new;
+  GdkColor *old;
 
   old = g_hash_table_lookup (target, name);
   if (!old || !gdk_color_equal (old, color))
     {
-      new = gdk_color_copy (color);
-      g_hash_table_insert (target, g_strdup (name), new);
+      g_hash_table_insert (target, g_strdup (name), gdk_color_copy (color));
       
       return TRUE;
     }
@@ -1699,7 +2016,7 @@ add_colors_to_hash_from_string (GHashTable  *hash,
 	}
 
       if (gdk_color_parse (p, &color))
-	changed = add_color_to_hash (name, &color, hash);
+	changed |= add_color_to_hash (name, &color, hash);
     }
 
   g_free (copy);
@@ -1747,7 +2064,8 @@ update_color_hash (ColorSchemeData   *data,
     return FALSE;
     
   /* Rebuild the merged hash table. */
-  g_hash_table_unref (data->color_hash);
+  if (data->color_hash)
+    g_hash_table_unref (data->color_hash);
   data->color_hash = g_hash_table_new_full (g_str_hash, g_str_equal, g_free,
 					    (GDestroyNotify) gdk_color_free);
   for (i = 0; i <= GTK_SETTINGS_SOURCE_APPLICATION; i++)
@@ -1760,46 +2078,27 @@ update_color_hash (ColorSchemeData   *data,
   return TRUE;
 }
 
-static void
-color_scheme_data_free (ColorSchemeData *data)
-{
-  gint i;
-
-  g_hash_table_unref (data->color_hash);
-
-  for (i = 0; i <= GTK_SETTINGS_SOURCE_APPLICATION; i++)
-    {
-      if (data->tables[i])
-	g_hash_table_unref (data->tables[i]);
-      g_free (data->lastentry[i]);
-    }
-
-  g_free (data);
-}
-
 static void
 merge_color_scheme (GtkSettings       *settings, 
-		    GValue            *value, 
+		    const GValue      *value, 
 		    GtkSettingsSource  source)
 {
   ColorSchemeData *data;
   const gchar *colors;
 
+  g_object_freeze_notify (G_OBJECT (settings));
+
   colors = g_value_get_string (value);
 
+  settings_update_color_scheme (settings);
+
   data = (ColorSchemeData *) g_object_get_data (G_OBJECT (settings),
 						"gtk-color-scheme");
-  if (!data)
-    {
-      data = g_new0 (ColorSchemeData, 1);
-      data->color_hash = g_hash_table_new_full (g_str_hash, g_str_equal, g_free, 
-						(GDestroyNotify) gdk_color_free);
-      g_object_set_data_full (G_OBJECT (settings), "gtk-color-scheme",
-			      data, (GDestroyNotify) color_scheme_data_free);
-    }
   
   if (update_color_hash (data, colors, source))
     g_object_notify (G_OBJECT (settings), "color-hash");
+
+  g_object_thaw_notify (G_OBJECT (settings));
 }
 
 static GHashTable *
@@ -1807,15 +2106,11 @@ get_color_hash (GtkSettings *settings)
 {
   ColorSchemeData *data;
 
-  if (!g_object_get_data (G_OBJECT (settings), "gtk-color-scheme"))
-    settings_update_color_scheme (settings);
+  settings_update_color_scheme (settings);
   
   data = (ColorSchemeData *)g_object_get_data (G_OBJECT (settings), 
 					       "gtk-color-scheme");
 
-  if (!data)
-    return NULL;
-
   return data->color_hash;
 }
 
@@ -1838,6 +2133,8 @@ get_color_scheme (GtkSettings *settings)
   ColorSchemeData *data;
   GString *string;
   
+  settings_update_color_scheme (settings);
+
   data = (ColorSchemeData *) g_object_get_data (G_OBJECT (settings),
 						"gtk-color-scheme");
 
